#!/usr/bin/env python
'''
A utility for easy manipulation of PNG files. 

Version 1.0
Dong Weiguo
'''

# for vim usage
# vim: tabstop=4 expandtab shiftwidth=4


import os
import sys
import argparse
import string
import struct 
import binascii
import zlib
#import array 
#import pylab
import mmap
import time     # for timing
hasLogging = True
try:
    import logging  # for better logging functionsa (requires Python 2.3+)
except ImportError:
    import mylogging as logging # a simple logging tool
    hasLogging = False

if hasLogging:
    # Get the logger. 
    logger = logging.getLogger("png")
    # By default, log to stdout
    stdout_hdlr = logging.StreamHandler( sys.stdout )
    formatter = logging.Formatter('%(asctime)s %(levelname)s %(message)s')
    stdout_hdlr.setFormatter( formatter )
    logger.addHandler( stdout_hdlr )
    logger.setLevel( logging.INFO )
else:
    logger = logging.MyLogging()

class Chunk(object):

    '''Definition of a chunk contained in a PNG file '''

    def __init__(self, buf):
        '''TODO: to be defined1. '''
        self.__length = 0
        self.__type = 0
        self.__data = []
        self.__crc = 0
        self.__size = 0     # The total size of the chunk in bytes

        if buf:
            self.load(buf)

    def load(self, buf):
        '''Load the chunk from the given buffer.

        @param buf List of bytes read from the PNG file
        @return: None

        '''
        length_buf = binascii.hexlify(buf[:4])
        #self.__length = int(''.join(length_buf), 16)
        self.__length = int(length_buf, 16)


        fmt = '4c'
        #self.__type = ''.join(struct.unpack(fmt, ''.join(buf[4:8])))
        self.__type = buf[4:8].decode()

        end_data = 8+self.__length
        self.__data = buf[8:end_data]
        
        buf = buf[end_data:]
        crc_bytes = binascii.hexlify(buf[:4])
        #self.__crc = int(''.join(crc_bytes), 16)
        self.__crc = int(crc_bytes, 16)

        self.__size = 8 + self.__length + 4;

    def totalSize(self):
        '''TODO: Returns the total size of the chunk in bytes.
        @return: TODO

        '''
        return self.__size

    def type(self):
        '''TODO: Docstring for type.
        @return: TODO

        '''
        return self.__type

    def data(self):
        '''TODO: Docstring for data.
        @return: TODO

        '''
        return self.__data

    def length(self):
        '''TODO: Docstring for length.
        @return: TODO

        '''
        return self.__length

class PNG(object):

    '''class for a PNG file. '''

    def __init__(self, file_path):
        '''TODO: to be defined1. '''
        self.__fileName = file_path
        self.__file = None
        self.__fobj = None
        self.size = [0, 0]
        self.__chunks = []  # List of chunks
        self.__width = 0
        self.__height = 0
        self.__bitDepth = 0
        self.__colorType = 0
        self.__cmpMtd = 0
        self.__fltMtd = 0
        self.__itlMtd = 0
        self.__totalIDAT = 0

        if file_path:
            self.readFile(file_path)
            self.__file = file_path
    
    def loadHeader(self, buf):
        '''TODO: Docstring for loadHeader.

        @param buf Byte buffer read from the import file
        @return: TODO

        '''
        pass

    def readFile(self, file_path):
        '''TODO: Docstring for readFile.

        @param file_path Full path to the png file
        @return: True if successful, False otherwise

        '''
        
        if file_path is None and self.__file is None:
          logger.info("File name is empty. Nothing to load!")
          return
          
        if file_path is not None:
          self.__file = file_path
        
        # Open the file for reading
        self.__fobj = open( self.__file, 'rb')

        fsize = os.path.getsize( file_path )
        # map the file into memory
        fno = self.__fobj.fileno()
        data = mmap.mmap( fno, fsize, access=mmap.ACCESS_READ )

        fmt = '8c'
        header = struct.unpack( fmt, data[:8] )

        # Get the header information
        #self.loadHeader() 
        #logger.debug("Header: %s" %  header )
        
        #working buffer   
        buf = data[8:]
        ihdr = Chunk(buf)
        self.parseIHDR(ihdr.data() )

        chunk_size = ihdr.totalSize()
        self.__chunks = [] 
        self.__chunks.append(ihdr)
        while len(buf) > chunk_size:
            buf = buf[chunk_size:]
            ck = Chunk(buf)
            chunk_size = ck.totalSize()
            self.__chunks.append(ck)
       
        self.parseIDAT()


        ### close the file
        data.close()
        self.__fobj.close()
        
    def parseIHDR(self, buf):
        '''TODO: Docstring for parseIHDR.

        @param buf The byte buffer of the chunk read from the file
        @return: TODO

        '''
        fmt = '!II5B'
        ihdr_vals = struct.unpack(fmt, buf)
        self.__width = ihdr_vals[0]
        self.__height = ihdr_vals[1]
        self.__bitDepth = ihdr_vals[2]
        self.__colorType = ihdr_vals[3]
        self.__cmpMtd = ihdr_vals[4]
        self.__fltMtd = ihdr_vals[5]
        self.__itlMtd = ihdr_vals[6]

    def parseIDAT(self):
        '''TODO: Parse IDAT chunks to get the total IDAT size.
        @return: TODO

        '''
        self.__totalIDAT = 0
        for i in range(1, len(self.__chunks)):
            if self.__chunks[i].type() == 'IDAT':
                self.__totalIDAT += self.__chunks[i].length()

    def testCompress(self):
        '''TODO: Compress the image data with various parameters using zlib.
        
        @return: None

        '''
        idat_data = b''
        self.__totalIDAT = 0
        for i in range(1, len(self.__chunks)):
            if self.__chunks[i].type() == 'IDAT':
                idat_data += self.__chunks[i].data()

        img_data = zlib.decompress( idat_data )
        print ('size of img_data = %d' % len(img_data) )
        fmt = '!BB'
        vals = struct.unpack(fmt, idat_data[:2])
        print ('zlib header in the PNG file = %x, %x' %(vals[0], vals[1]) )

        zlib_strategy = (0, 1)
        zlib_memlevel = range(1, 10)    # valid memory level between 1 and 9
        zlib_compresslevel = range(10)  # compress level between 0 and 9

        msg_buf = ''
        found_cmp = False

        for zs in zlib_strategy:
            for zm in zlib_memlevel:
                for zc in zlib_compresslevel:
                    cmp_level = zc
                    BLOCK_SIZE=8192
                    cmp_obj = zlib.compressobj(cmp_level, zlib.DEFLATED, 15, zm, zs)
                    cmp_tmp = cmp_obj.compress(img_data)

                    if False:
                        img_tmp = img_data
                        cmp_tmp = cmp_obj.compress(img_tmp[:BLOCK_SIZE])
                        img_tmp = img_tmp[BLOCK_SIZE:]
                        while len(img_tmp)>=BLOCK_SIZE:
                            cmp_tmp += cmp_obj.compress(img_tmp[:BLOCK_SIZE])
                            img_tmp = img_tmp[BLOCK_SIZE:]
                        cmp_tmp += cmp_obj.compress(img_tmp)

                    cmp_tmp += cmp_obj.flush()
                    if len(cmp_tmp) == len(idat_data):
                        print( 'With compression level %d, mem_level %d, strategy %d, zlib.compress gives %d bytes' % (cmp_level, zm, zs, len(cmp_tmp)))
                        found_cmp = True
                        break
                    else:
                        msg_buf += 'With compression level %d, mem_level %d, strategy %d, zlib.compress gives %d bytes\n' % (cmp_level, zm, zs, len(cmp_tmp))
                    #vals = struct.unpack('!BB', cmp_tmp[:2])
                    #print ('Python zlib header = %x, %x' %(vals[0], vals[1]) )

        if not found_cmp:
            print(msg_buf)

    def dump(self):
        '''TODO: Docstring for function.

        @return: TODO

        '''
        print('\n\nInformation about %s' % self.__fileName)
        print('IHDR information:')
        print('\tWidth: %d'% self.__width)
        print('\tHeight: %d'%self.__height)
        print('\tBit depth: %d' % self.__bitDepth)
        print('\tColor type: %d' % self.__colorType)
        print('\tCompression method: %d'% self.__cmpMtd)
        print('\tFilter method: %d'% self.__fltMtd)
        print('\tInterlace method: %d'% self.__itlMtd)

        first_idat = True
        for i in range(1, len(self.__chunks)):
            ck = self.__chunks[i]
            if ck.type() == 'IDAT':
                if first_idat:
                    print('\n Chunk %s: IDAT chunk size: %d bytes, total IDAT size: %d bytes'%(ck.type(),ck.length(), self.__totalIDAT))
                    first_idat = False
                continue
                
            print('\n Chunk %s: Data length: %d, total size: %d bytes'%(ck.type(), ck.length(), ck.totalSize()))


def main():

    parser = argparse.ArgumentParser()
    parser.add_argument('pngfile', help='the given png file name for analysis')
    args = parser.parse_args()
    
    pf = PNG(args.pngfile)
    pf.dump()
    pf.testCompress()

if __name__ == "__main__":
    main()

